try:
    import simplejson as json
except ImportError:
    raise ImportError("`simplejson` not installed")
from typing import List, Optional, Dict, Any, Union

from phi.tools import Toolkit
from phi.utils.log import logger

try:
    from sqlalchemy import create_engine, Engine, Row
    from sqlalchemy.orm import Session, sessionmaker
    from sqlalchemy.inspection import inspect
    from sqlalchemy.sql.expression import text
except ImportError:
    raise ImportError("`sqlalchemy` not installed")


class SQLToolkit(Toolkit):
    def __init__(
        self,
        db_url: Optional[str] = None,
        db_engine: Optional[Engine] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        schema: Optional[str] = None,
        dialect: Optional[str] = None,
        tables: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(name="sql_toolkit")

        # Get the database engine
        _engine: Optional[Engine] = db_engine
        if _engine is None and db_url is not None:
            _engine = create_engine(db_url)
        elif user and password and host and port and dialect:
            if schema is not None:
                _engine = create_engine(f"{dialect}://{user}:{password}@{host}:{port}/{schema}")
            else:
                _engine = create_engine(f"{dialect}://{user}:{password}@{host}:{port}")

        if _engine is None:
            raise ValueError("Could not build the database connection")

        # Database connection
        self.db_engine: Engine = _engine
        self.Session: sessionmaker[Session] = sessionmaker(bind=self.db_engine)

        # Tables this toolkit can access
        self.tables: Optional[Dict[str, Any]] = tables

        # Register functions in the toolkit
        self.register(self.run_sql_query)
        self.register(self.run_sql_query_and_get_result)
        self.register(self.get_table_names)
        self.register(self.describe_table)

    def get_table_names(self) -> str:
        """
        Use this function to get a list of table names you have access to.

        Returns:
            str: list of tables in the database.
        """
        if self.tables is not None:
            return json.dumps(self.tables)

        try:
            table_names = inspect(self.db_engine).get_table_names()
            logger.debug(f"table_names: {table_names}")
        except Exception as e:
            logger.error(f"Error getting tables: {e}")
            return f"Error getting tables: {e}"

        return json.dumps(table_names)

    def describe_table(self, table_name: str) -> str:
        """
        Use this function to get the schema of an existing table.

        Args:
            table_name (str): The name of the table to get the schema for.

        Returns:
            str: schema of a table
        """

        try:
            table_names = inspect(self.db_engine)
            table_schema = table_names.get_columns(table_name)

        except Exception as e:
            logger.error(f"Error getting table schema: {e}")
            return f"Error getting table schema: {e}"

        return json.dumps([str(column) for column in table_schema])

    def run_sql_query(self, query: str) -> str:
        """
        Use this function to run a SQL query, it does not return any output.

        Args:
            query (str): The query to run.

        Returns:
            str: Result of the SQL query.
        """

        try:
            return json.dumps(self._run_sql(sql=query, output=False))
        except Exception as e:
            logger.error(f"Error running query: {e}")
            return f"Error running query: {e}"

    def run_sql_query_and_get_result(self, query: str, limit: Optional[int] = 10) -> str:
        """
        Use this function to run a SQL query, it returns output of the query.

        Args:
            query (str): The query to run.
            limit (int, optional): The number of rows to return. Defaults to 10. Use `None` to show all results.

        Returns:
            str: Result of the SQL query.
        """
        try:
            return json.dumps(str(self._run_sql(sql=query, limit=limit, output=True)))
        except Exception as e:
            logger.error(f"Error running query: {e}")
            return f"Error running query: {e}"

    def _run_sql(
        self, sql: str, limit: Optional[int] = None, output: Optional[bool] = None
    ) -> Optional[Union[List, Dict]]:
        """Internal function to run a sql query.

        Args:
            sql (str): The sql query to run.
            limit (int, optional): The number of rows to return. Defaults to None.
            output (bool, optional): Whether the query requires an output. Defaults to None.

        Returns:
            List[dict]: The result of the query.
        """
        logger.debug(f"Running sql |\n{sql}")

        with self.Session.begin() as session:
            if output:
                if limit is None:
                    sql_result = session.execute(text(sql)).fetchall()
                else:
                    sql_result = session.execute(text(sql)).fetchmany(limit)
            else:
                sql_result = session.execute(text(sql))  # type: ignore

        logger.debug(f"SQL result: {sql_result}")
        if sql_result is None:
            return None
        elif isinstance(sql_result, list):
            return [row._asdict() for row in sql_result]
        elif isinstance(sql_result, Row):
            return sql_result._asdict()
        else:
            logger.debug(f"SQL result type: {type(sql_result)}")
            return None



import json
from os import getenv
from typing import Optional

from phi.tools import Toolkit
from phi.utils.log import logger

try:
    from exa_py import Exa
except ImportError:
    raise ImportError("`exa_py` not installed. Please install using `pip install exa_py`")


class ExaTools(Toolkit):
    def __init__(
        self,
        api_key: Optional[str] = None,
        search: bool = False,
        search_with_contents: bool = True,
        show_results: bool = False,
    ):
        super().__init__(name="exa")

        self.api_key = api_key or getenv("EXA_API_KEY")
        if not self.api_key:
            logger.error("No Exa API key provided")

        self.show_results = show_results
        if search:
            self.register(self.search_exa)
        if search_with_contents:
            self.register(self.search_exa_with_contents)

    def search_exa(self, query: str, num_results: int = 5) -> str:
        """Searches Exa for a query.

        :param query: The query to search for.
        :param num_results: The number of results to return.
        :return: Links of relevant documents from exa.
        """
        if not self.api_key:
            return "Please provide the EXA_API_KEY"

        try:
            exa = Exa(self.api_key)
            logger.debug(f"Searching exa for: {query}")
            exa_results = exa.search(query, num_results=num_results)
            exa_search_urls = [result.url for result in exa_results.results]
            parsed_results = "\n".join(exa_search_urls)
            if self.show_results:
                logger.info(parsed_results)
            return parsed_results
        except Exception as e:
            logger.error(f"Failed to search exa {e}")
            return f"Error: {e}"

    def search_exa_with_contents(self, query: str, num_results: int = 3, text_length_limit: int = 1000) -> str:
        """Searches Exa for a query and returns the contents from the search results.

        :param query: The query to search for.
        :param num_results: The number of results to return. Defaults to 3.
        :param text_length_limit: The length of the text to return. Defaults to 1000.
        :return: JSON string of the search results.
        """
        if not self.api_key:
            return "Please provide the EXA_API_KEY"

        try:
            exa = Exa(self.api_key)
            logger.debug(f"Searching exa for: {query}")
            exa_results = exa.search_and_contents(query, num_results=num_results)
            exa_results_parsed = []
            for result in exa_results.results:
                result_dict = {"url": result.url}
                if result.text:
                    result_dict["text"] = result.text[:text_length_limit]
                if result.author:
                    result_dict["author"] = result.author
                if result.title:
                    result_dict["title"] = result.title
                exa_results_parsed.append(result_dict)

            parsed_results = json.dumps(exa_results_parsed, indent=2)
            if self.show_results:
                logger.info(parsed_results)
            return parsed_results
        except Exception as e:
            logger.error(f"Failed to search exa {e}")
            return f"Error: {e}"



import json

from phi.tools import Toolkit

try:
    import yfinance as yf
except ImportError:
    raise ImportError("`yfinance` not installed. Please install using `pip install yfinance`.")


class YFinanceTools(Toolkit):
    def __init__(
        self,
        stock_price: bool = True,
        stock_fundamentals: bool = False,
        income_statements: bool = False,
        key_financial_ratios: bool = False,
        analyst_recommendations: bool = False,
        company_news: bool = False,
        technical_indicators: bool = False,
        company_profile: bool = False,
    ):
        super().__init__(name="yfinance_tools")

        if stock_price:
            self.register(self.get_current_stock_price)
        if stock_fundamentals:
            self.register(self.get_stock_fundamentals)
        if income_statements:
            self.register(self.get_income_statements)
        if key_financial_ratios:
            self.register(self.get_key_financial_ratios)
        if analyst_recommendations:
            self.register(self.get_analyst_recommendations)
        if company_news:
            self.register(self.get_company_news)
        if technical_indicators:
            self.register(self.get_technical_indicators)
        if company_profile:
            self.register(self.get_company_profile)

    def get_current_stock_price(self, symbol: str) -> str:
        """
        Get the current stock price for a given symbol.

        Args:
          symbol (str): The stock symbol.

        Returns:
          str: The current stock price or error message.
        """
        try:
            stock = yf.Ticker(symbol)
            # Use "regularMarketPrice" for regular market hours, or "currentPrice" for pre/post market
            current_price = stock.info.get("regularMarketPrice", stock.info.get("currentPrice"))
            return f"{current_price:.4f}" if current_price else f"Could not fetch current price for {symbol}"
        except Exception as e:
            return f"Error fetching current price for {symbol}: {e}"

    def get_stock_fundamentals(self, symbol: str) -> str:
        """
        Get fundamental data for a given stock symbol yfinance API.

        Args:
            symbol (str): The stock symbol.

        Returns:
            str: A JSON string containing fundamental data or an error message.
                Keys:
                    - 'symbol': The stock symbol.
                    - 'company_name': The long name of the company.
                    - 'sector': The sector to which the company belongs.
                    - 'industry': The industry to which the company belongs.
                    - 'market_cap': The market capitalization of the company.
                    - 'pe_ratio': The forward price-to-earnings ratio.
                    - 'pb_ratio': The price-to-book ratio.
                    - 'dividend_yield': The dividend yield.
                    - 'eps': The trailing earnings per share.
                    - 'beta': The beta value of the stock.
                    - '52_week_high': The 52-week high price of the stock.
                    - '52_week_low': The 52-week low price of the stock.
        """
        try:
            stock = yf.Ticker(symbol)
            info = stock.info
            fundamentals = {
                "symbol": symbol,
                "company_name": info.get("longName", ""),
                "sector": info.get("sector", ""),
                "industry": info.get("industry", ""),
                "market_cap": info.get("marketCap", "N/A"),
                "pe_ratio": info.get("forwardPE", "N/A"),
                "pb_ratio": info.get("priceToBook", "N/A"),
                "dividend_yield": info.get("dividendYield", "N/A"),
                "eps": info.get("trailingEps", "N/A"),
                "beta": info.get("beta", "N/A"),
                "52_week_high": info.get("fiftyTwoWeekHigh", "N/A"),
                "52_week_low": info.get("fiftyTwoWeekLow", "N/A"),
            }
            return json.dumps(fundamentals, indent=2)
        except Exception as e:
            return f"Error getting fundamentals for {symbol}: {e}"

    def get_income_statements(self, symbol: str) -> str:
        """
        Get income statements for a given stock symbol.

        Args:
        symbol (str): The stock symbol.

        Returns:
        dict: JSON containing income statements or an empty dictionary.
        """
        try:
            stock = yf.Ticker(symbol)
            financials = stock.financials
            return financials.to_json(orient="index")
        except Exception as e:
            return f"Error fetching income statements for {symbol}: {e}"

    def get_key_financial_ratios(self, symbol: str) -> str:
        """
        Get key financial ratios for a given stock symbol.

        Args:
        symbol (str): The stock symbol.

        Returns:
        dict: JSON containing key financial ratios.
        """
        try:
            stock = yf.Ticker(symbol)
            key_ratios = stock.info
            return json.dumps(key_ratios, indent=2)
        except Exception as e:
            return f"Error fetching key financial ratios for {symbol}: {e}"

    def get_analyst_recommendations(self, symbol: str) -> str:
        """
        Get analyst recommendations for a given stock symbol.

        Args:
        symbol (str): The stock symbol.

        Returns:
        str: JSON containing analyst recommendations.
        """
        try:
            stock = yf.Ticker(symbol)
            recommendations = stock.recommendations
            return recommendations.to_json(orient="index")
        except Exception as e:
            return f"Error fetching analyst recommendations for {symbol}: {e}"

    def get_company_news(self, symbol: str, num_stories: int = 3) -> str:
        """
        Get company news and press releases for a given stock symbol.

        Args:
        symbol (str): The stock symbol.
        num_stories (int): The number of news stories to return. Defaults to 3.

        Returns:
        str: JSON containing company news and press releases.
        """
        try:
            news = yf.Ticker(symbol).news
            return json.dumps(news[:num_stories], indent=2)
        except Exception as e:
            return f"Error fetching company news for {symbol}: {e}"

    def get_technical_indicators(self, symbol: str, period: str = "3mo") -> str:
        """
        Get technical indicators for a given stock symbol.

        Args:
        symbol (str): The stock symbol.
        period (str): The time period for which to retrieve technical indicators.
            Valid periods: 1d, 5d, 1mo, 3mo, 6mo, 1y, 2y, 5y, 10y, ytd, max. Defaults to 3mo.

        Returns:
        str: JSON containing technical indicators.
        """
        try:
            indicators = yf.Ticker(symbol).history(period=period)
            return indicators.to_json(orient="index")
        except Exception as e:
            return f"Error fetching technical indicators for {symbol}: {e}"

    def get_company_profile(self, symbol: str) -> str:
        """
        Get company profile and overview for a given stock symbol.

        Args:
        symbol (str): The stock symbol.

        Returns:
        str: JSON containing company profile and overview.
        """
        try:
            profile = yf.Ticker(symbol).info
            return json.dumps(profile, indent=2)
        except Exception as e:
            return f"Error fetching company profile for {symbol}: {e}"



from typing import List
from subprocess import run, CompletedProcess

from phi.cli.console import print_info
from phi.utils.log import logger


def run_shell_command(args: List[str], display_result: bool = True, display_error: bool = True) -> CompletedProcess:
    logger.debug(f"Running command: {args}")
    result = run(args, capture_output=True, text=True)
    if result.returncode != 0:
        raise Exception(result.stderr)
    if result.stdout and display_result:
        print_info(result.stdout)
    if result.stderr and display_error:
        print_info(result.stderr)
    return result



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import FilePath

from phi.resource.base import ResourceBase
from phi.k8s.api_client import K8sApiClient
from phi.k8s.constants import DEFAULT_K8S_NAMESPACE
from phi.k8s.helm.cli import run_shell_command
from phi.cli.console import print_info
from phi.utils.log import logger


class HelmChart(ResourceBase):
    chart: str
    set: Optional[Dict[str, Any]] = None
    values: Optional[Union[FilePath, List[FilePath]]] = None
    flags: Optional[List[str]] = None
    namespace: Optional[str] = None
    create_namespace: bool = True

    repo: Optional[str] = None
    update_repo_before_install: bool = True

    k8s_client: Optional[K8sApiClient] = None
    resource_type: str = "Chart"

    def get_resource_name(self) -> str:
        return self.name

    def get_namespace(self) -> str:
        if self.namespace is not None:
            return self.namespace
        return DEFAULT_K8S_NAMESPACE

    def get_k8s_client(self) -> K8sApiClient:
        if self.k8s_client is not None:
            return self.k8s_client
        self.k8s_client = K8sApiClient()
        return self.k8s_client

    def _read(self, k8s_client: K8sApiClient) -> Any:
        try:
            logger.info(f"Getting helm chart: {self.name}\n")
            get_args = ["helm", "get", "manifest", self.name]
            if self.namespace is not None:
                get_args.append(f"--namespace={self.namespace}")
            get_result = run_shell_command(get_args, display_result=False, display_error=False)
            if get_result.stdout:
                import yaml

                return yaml.safe_load_all(get_result.stdout)
        except Exception:
            pass
        return None

    def read(self, k8s_client: K8sApiClient) -> Any:
        # Step 1: Use cached value if available
        if self.use_cache and self.active_resource is not None:
            return self.active_resource

        # Step 2: Skip resource creation if skip_read = True
        if self.skip_read:
            print_info(f"Skipping read: {self.get_resource_name()}")
            return True

        # Step 3: Read resource
        client: K8sApiClient = k8s_client or self.get_k8s_client()
        return self._read(client)

    def is_active(self, k8s_client: K8sApiClient) -> bool:
        """Returns True if the resource is active on the k8s cluster"""
        self.active_resource = self._read(k8s_client=k8s_client)
        return True if self.active_resource is not None else False

    def _create(self, k8s_client: K8sApiClient) -> bool:
        if self.repo:
            try:
                logger.info(f"Adding helm repo: {self.name} {self.repo}\n")
                add_args = ["helm", "repo", "add", self.name, self.repo]
                run_shell_command(add_args)

                if self.update_repo_before_install:
                    logger.info(f"Updating helm repo: {self.name}\n")
                    update_args = ["helm", "repo", "update", self.name]
                    run_shell_command(update_args)
            except Exception as e:
                logger.error(f"Failed to add helm repo: {e}")
                return False

        try:
            logger.info(f"Installing helm chart: {self.name}\n")
            install_args = ["helm", "install", self.name, self.chart]
            if self.set is not None:
                for key, value in self.set.items():
                    install_args.append(f"--set {key}={value}")
            if self.flags:
                install_args.extend(self.flags)
            if self.values:
                if isinstance(self.values, Path):
                    install_args.append(f"--values={str(self.values)}")
                elif isinstance(self.values, list):
                    for value in self.values:
                        install_args.append(f"--values={str(value)}")
            if self.namespace is not None:
                install_args.append(f"--namespace={self.namespace}")
                if self.create_namespace:
                    install_args.append("--create-namespace")
            run_shell_command(install_args)
            return True
        except Exception as e:
            logger.error(f"Failed to install helm chart: {e}")
        return False

    def create(self, k8s_client: K8sApiClient) -> bool:
        # Step 1: Skip resource creation if skip_create = True
        if self.skip_create:
            print_info(f"Skipping create: {self.get_resource_name()}")
            return True

        # Step 2: Check if resource is active and use_cache = True
        client: K8sApiClient = k8s_client or self.get_k8s_client()
        if self.use_cache and self.is_active(client):
            self.resource_created = True
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} already exists")
            return True

        # Step 3: Create the resource
        else:
            self.resource_created = self._create(client)
            if self.resource_created:
                print_info(f"{self.get_resource_type()}: {self.get_resource_name()} created")

        # Step 4: Run post create steps
        if self.resource_created:
            if self.save_output:
                self.save_output_file()
            logger.debug(f"Running post-create for {self.get_resource_type()}: {self.get_resource_name()}")
            return self.post_create(client)
        logger.error(f"Failed to create {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_created

    def post_create(self, k8s_client: K8sApiClient) -> bool:
        return True

    def _update(self, k8s_client: K8sApiClient) -> Any:
        try:
            logger.info(f"Updating helm chart: {self.name}\n")
            update_args = ["helm", "upgrade", self.name, self.chart]
            if self.set is not None:
                for key, value in self.set.items():
                    update_args.append(f"--set {key}={value}")
            if self.flags:
                update_args.extend(self.flags)
            if self.values:
                if isinstance(self.values, Path):
                    update_args.append(f"--values={str(self.values)}")
            if self.namespace is not None:
                update_args.append(f"--namespace={self.namespace}")
            run_shell_command(update_args)
            return True
        except Exception as e:
            logger.error(f"Failed to update helm chart: {e}")
        return False

    def update(self, k8s_client: K8sApiClient) -> bool:
        # Step 1: Skip resource update if skip_update = True
        if self.skip_update:
            print_info(f"Skipping update: {self.get_resource_name()}")
            return True

        # Step 2: Update the resource
        client: K8sApiClient = k8s_client or self.get_k8s_client()
        if self.is_active(client):
            self.resource_updated = self._update(client)
        else:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} does not exist")
            return True

        # Step 3: Run post update steps
        if self.resource_updated:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} updated")
            if self.save_output:
                self.save_output_file()
            logger.debug(f"Running post-update for {self.get_resource_type()}: {self.get_resource_name()}")
            return self.post_update(client)
        logger.error(f"Failed to update {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_updated

    def post_update(self, k8s_client: K8sApiClient) -> bool:
        return True

    def _delete(self, k8s_client: K8sApiClient) -> Any:
        try:
            logger.info(f"Deleting helm chart: {self.name}\n")
            delete_args = ["helm", "uninstall", self.name]
            if self.namespace is not None:
                delete_args.append(f"--namespace={self.namespace}")
            run_shell_command(delete_args)
            return True
        except Exception as e:
            logger.error(f"Failed to delete helm chart: {e}")
        return False

    def delete(self, k8s_client: K8sApiClient) -> bool:
        # Step 1: Skip resource deletion if skip_delete = True
        if self.skip_delete:
            print_info(f"Skipping delete: {self.get_resource_name()}")
            return True

        # Step 2: Delete the resource
        client: K8sApiClient = k8s_client or self.get_k8s_client()
        if self.is_active(client):
            self.resource_deleted = self._delete(client)
        else:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} does not exist")
            return True

        # Step 3: Run post delete steps
        if self.resource_deleted:
            print_info(f"{self.get_resource_type()}: {self.get_resource_name()} deleted")
            if self.save_output:
                self.delete_output_file()
            logger.debug(f"Running post-delete for {self.get_resource_type()}: {self.get_resource_name()}.")
            return self.post_delete(client)
        logger.error(f"Failed to delete {self.get_resource_type()}: {self.get_resource_name()}")
        return self.resource_deleted

    def post_delete(self, k8s_client: K8sApiClient) -> bool:
        return True



import json
from uuid import uuid4
from typing import List, Any, Optional, Dict, Iterator, Callable, Union, Type, Tuple, Literal

from pydantic import BaseModel, ConfigDict, field_validator, Field, ValidationError

from phi.assistant.run import AssistantRun
from phi.knowledge.base import AssistantKnowledge
from phi.llm.base import LLM
from phi.llm.message import Message
from phi.llm.references import References  # noqa: F401
from phi.memory.assistant import AssistantMemory
from phi.prompt.template import PromptTemplate
from phi.storage.assistant import AssistantStorage
from phi.task.task import Task
from phi.task.llm import LLMTask
from phi.tools import Tool, Toolkit, Function
from phi.utils.log import logger, set_log_level_to_debug
from phi.utils.message import get_text_from_message
from phi.utils.merge_dict import merge_dictionaries
from phi.utils.timer import Timer


class Assistant(BaseModel):
    # -*- Assistant settings
    # LLM to use for this Assistant
    llm: Optional[LLM] = None
    # Assistant introduction. This is added to the chat history when a run is started.
    introduction: Optional[str] = None
    # Assistant name
    name: Optional[str] = None
    # Metadata associated with this assistant
    assistant_data: Optional[Dict[str, Any]] = None

    # -*- Run settings
    # Run UUID (autogenerated if not set)
    run_id: Optional[str] = Field(None, validate_default=True)
    # Run name
    run_name: Optional[str] = None
    # Metadata associated with this run
    run_data: Optional[Dict[str, Any]] = None

    # -*- User settings
    # ID of the user participating in this run
    user_id: Optional[str] = None
    # Metadata associated the user participating in this run
    user_data: Optional[Dict[str, Any]] = None

    # -*- Assistant Memory
    memory: AssistantMemory = AssistantMemory()
    # add_chat_history_to_messages=true_adds_the_chat_history_to_the_messages_sent_to_the_llm.
    add_chat_history_to_messages: bool = False
    # add_chat_history_to_prompt=True adds the formatted chat history to the user prompt.
    add_chat_history_to_prompt: bool = False
    # Number of previous messages to add to the prompt or messages.
    num_history_messages: int = 6

    # -*- Assistant Knowledge Base
    knowledge_base: Optional[AssistantKnowledge] = None
    # Enable RAG by adding references from the knowledge base to the prompt.
    add_references_to_prompt: bool = False

    # -*- Assistant Storage
    storage: Optional[AssistantStorage] = None
    # AssistantRun from the database: DO NOT SET MANUALLY
    db_row: Optional[AssistantRun] = None
    # -*- Assistant Tools
    # A list of tools provided to the LLM.
    # Tools are functions the model may generate JSON inputs for.
    # If you provide a dict, it is not called by the model.
    tools: Optional[List[Union[Tool, Toolkit, Callable, Dict, Function]]] = None
    # Allow the assistant to use tools
    use_tools: bool = False
    # Show tool calls in LLM messages.
    show_tool_calls: bool = False
    # Maximum number of tool calls allowed.
    tool_call_limit: Optional[int] = None
    # Controls which (if any) tool is called by the model.
    # "none" means the model will not call a tool and instead generates a message.
    # "auto" means the model can pick between generating a message or calling a tool.
    # Specifying a particular function via {"type: "function", "function": {"name": "my_function"}}
    #   forces the model to call that tool.
    # "none" is the default when no tools are present. "auto" is the default if tools are present.
    tool_choice: Optional[Union[str, Dict[str, Any]]] = None
    # -*- Available tools
    # If use_tools is True and read_chat_history_tool is True,
    # then a tool is added that allows the LLM to read the chat history.
    read_chat_history_tool: bool = True
    # If use_tools is True and search_knowledge_base_tool is True,
    # then a tool is added that allows the LLM to search the knowledge base.
    search_knowledge_base_tool: bool = True
    # If use_tools is True and update_knowledge_base is True,
    # then a tool is added that allows the LLM to update the knowledge base.
    update_knowledge_base: bool = False
    # If use_tools is True and read_tool_call_history is True,
    # then a tool is added that allows the LLM to get the tool call history.
    read_tool_call_history: bool = False

    # -*- Important: this setting determines if the input messages are formatted
    # If True, phidata will add the system prompt, references, and chat history
    # If False, the input messages are sent to the LLM as is
    format_messages: bool = True

    #
    # -*- Prompt Settings
    #
    # -*- System prompt: provide the system prompt as a string
    system_prompt: Optional[str] = None
    # -*- System prompt template: provide the system prompt as a PromptTemplate
    system_prompt_template: Optional[PromptTemplate] = None
    # -*- System prompt function: provide the system prompt as a function
    # This function is provided the "Assistant object" as an argument
    #   and should return the system_prompt as a string.
    # Signature:
    # def system_prompt_function(assistant: Assistant) -> str:
    #    ...
    system_prompt_function: Optional[Callable[..., Optional[str]]] = None
    # If True, build a default system prompt using instructions and extra_instructions
    build_default_system_prompt: bool = True
    # -*- Settings for building the default system prompt
    # Assistant description for the default system prompt
    description: Optional[str] = None
    # List of instructions for the default system prompt
    instructions: Optional[List[str]] = None
    # List of extra_instructions added to the default system prompt
    # Use these when you want to use the default prompt but also add some extra instructions
    extra_instructions: Optional[List[str]] = None
    # Add a string to the end of the default system prompt
    add_to_system_prompt: Optional[str] = None
    # If True, add instructions for using the knowledge base to the default system prompt if knowledge base is provided
    add_knowledge_base_instructions: bool = True
    # If True, add instructions for letting the user know that the assistant does not know the answer
    prevent_hallucinations: bool = False
    # If True, add instructions to prevent prompt injection attacks
    prevent_prompt_injection: bool = False
    # If True, add instructions for limiting tool access to the default system prompt if tools are provided
    limit_tool_access: bool = False
    # If True, add the current datetime to the prompt to give the assistant a sense of time
    # This allows for relative times like "tomorrow" to be used in the prompt
    add_datetime_to_instructions: bool = False
    # If markdown=true, add instructions to format the output using markdown
    markdown: bool = False

    # -*- User prompt: provide the user prompt as a string
    # Note: this will ignore the input message provided to the run function
    user_prompt: Optional[Union[List, Dict, str]] = None
    # -*- User prompt template: provide the user prompt as a PromptTemplate
    user_prompt_template: Optional[PromptTemplate] = None
    # -*- User prompt function: provide the user prompt as a function.
    # This function is provided the "Assistant object" and the "input message" as arguments
    #   and should return the user_prompt as a Union[List, Dict, str].
    # If add_references_to_prompt is True, then references are also provided as an argument.
    # If add_chat_history_to_prompt is True, then chat_history is also provided as an argument.
    # Signature:
    # def custom_user_prompt_function(
    #     assistant: Assistant,
    #     message: Union[List, Dict, str],
    #     references: Optional[str] = None,
    #     chat_history: Optional[str] = None,
    # ) -> Union[List, Dict, str]:
    #     ...
    user_prompt_function: Optional[Callable[..., str]] = None
    # If True, build a default user prompt using references and chat history
    build_default_user_prompt: bool = True
    # Function to get references for the user_prompt
    # This function, if provided, is called when add_references_to_prompt is True
    # Signature:
    # def references(assistant: Assistant, query: str) -> Optional[str]:
    #     ...
    references_function: Optional[Callable[..., Optional[str]]] = None
    references_format: Literal["json", "yaml"] = "json"
    # Function to get the chat_history for the user prompt
    # This function, if provided, is called when add_chat_history_to_prompt is True
    # Signature:
    # def chat_history(assistant: Assistant) -> str:
    #     ...
    chat_history_function: Optional[Callable[..., Optional[str]]] = None

    # -*- Assistant Output Settings
    # Provide an output model for the responses
    output_model: Optional[Union[str, List, Type[BaseModel]]] = None
    # If True, the output is converted into the output_model (pydantic model or json dict)
    parse_output: bool = True
    # -*- Final LLM response i.e. the final output of this assistant
    output: Optional[Any] = None

    # -*- Assistant Tasks
    # Tasks allow the Assistant to generate a response using a list of tasks
    # If tasks is None or empty, a single default LLM task is created for this assistant
    tasks: Optional[List[Task]] = None
    # Metadata associated with the assistant tasks
    task_data: Optional[Dict[str, Any]] = None

    # -*- Team settings
    role: Optional[str] = None

    # debug_mode=True enables debug logs
    debug_mode: bool = False
    # monitoring=True logs Assistant runs on phidata.com
    monitoring: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("debug_mode", mode="before")
    def set_log_level(cls, v: bool) -> bool:
        if v:
            set_log_level_to_debug()
            logger.debug("Debug logs enabled")
        return v

    @field_validator("run_id", mode="before")
    def set_run_id(cls, v: Optional[str]) -> str:
        return v if v is not None else str(uuid4())

    @property
    def streamable(self) -> bool:
        return self.output_model is None

    @property
    def llm_task(self) -> LLMTask:
        """Returns an LLMTask for this assistant"""

        _llm_task = LLMTask(
            llm=self.llm.model_copy() if self.llm is not None else None,
            assistant_name=self.name,
            assistant_memory=self.memory,
            add_references_to_prompt=self.add_references_to_prompt,
            add_chat_history_to_messages=self.add_chat_history_to_messages,
            num_history_messages=self.num_history_messages,
            knowledge_base=self.knowledge_base,
            use_tools=self.use_tools,
            show_tool_calls=self.show_tool_calls,
            tool_call_limit=self.tool_call_limit,
            tools=self.tools,
            tool_choice=self.tool_choice,
            read_chat_history_tool=self.read_chat_history_tool,
            search_knowledge_base_tool=self.search_knowledge_base_tool,
            update_knowledge_base=self.update_knowledge_base,
            read_tool_call_history=self.read_tool_call_history,
            format_messages=self.format_messages,
            system_prompt=self.system_prompt,
            system_prompt_template=self.system_prompt_template,
            system_prompt_function=self.system_prompt_function,
            build_default_system_prompt=self.build_default_system_prompt,
            description=self.description,
            instructions=self.instructions,
            extra_instructions=self.extra_instructions,
            add_to_system_prompt=self.add_to_system_prompt,
            add_knowledge_base_instructions=self.add_knowledge_base_instructions,
            prevent_hallucinations=self.prevent_hallucinations,
            prevent_prompt_injection=self.prevent_prompt_injection,
            limit_tool_access=self.limit_tool_access,
            add_datetime_to_instructions=self.add_datetime_to_instructions,
            markdown=self.markdown,
            user_prompt=self.user_prompt,
            user_prompt_template=self.user_prompt_template,
            user_prompt_function=self.user_prompt_function,
            build_default_user_prompt=self.build_default_user_prompt,
            references_function=self.references_function,
            references_format=self.references_format,
            chat_history_function=self.chat_history_function,
            output_model=self.output_model,
        )
        return _llm_task

    def to_database_row(self) -> AssistantRun:
        """Create a AssistantRun for the current Assistant (to save to the database)"""

        return AssistantRun(
            name=self.name,
            run_id=self.run_id,
            run_name=self.run_name,
            user_id=self.user_id,
            llm=self.llm.to_dict() if self.llm is not None else None,
            memory=self.memory.to_dict(),
            assistant_data=self.assistant_data,
            run_data=self.run_data,
            user_data=self.user_data,
            task_data=self.task_data,
        )

    def from_database_row(self, row: AssistantRun):
        """Load the existing Assistant from an AssistantRun (from the database)"""

        # Values that are overwritten from the database if they are not set in the assistant
        if self.name is None and row.name is not None:
            self.name = row.name
        if self.run_id is None and row.run_id is not None:
            self.run_id = row.run_id
        if self.run_name is None and row.run_name is not None:
            self.run_name = row.run_name
        if self.user_id is None and row.user_id is not None:
            self.user_id = row.user_id

        # Update llm data from the AssistantRun
        if row.llm is not None:
            # Update llm metrics from the database
            llm_metrics_from_db = row.llm.get("metrics")
            if llm_metrics_from_db is not None and isinstance(llm_metrics_from_db, dict) and self.llm:
                try:
                    self.llm.metrics = llm_metrics_from_db
                except Exception as e:
                    logger.warning(f"Failed to load llm metrics: {e}")

        # Update assistant memory from the AssistantRun
        if row.memory is not None:
            try:
                self.memory = self.memory.__class__.model_validate(row.memory)
            except Exception as e:
                logger.warning(f"Failed to load assistant memory: {e}")

        # Update assistant_data from the database
        if row.assistant_data is not None:
            # If assistant_data is set in the assistant, merge it with the database assistant_data.
            # The assistant assistant_data takes precedence
            if self.assistant_data is not None and row.assistant_data is not None:
                # Updates db_row.assistant_data with self.assistant_data
                merge_dictionaries(row.assistant_data, self.assistant_data)
                self.assistant_data = row.assistant_data
            # If assistant_data is not set in the assistant, use the database assistant_data
            if self.assistant_data is None and row.assistant_data is not None:
                self.assistant_data = row.assistant_data

        # Update run_data from the database
        if row.run_data is not None:
            # If run_data is set in the assistant, merge it with the database run_data.
            # The assistant run_data takes precedence
            if self.run_data is not None and row.run_data is not None:
                # Updates db_row.run_data with self.run_data
                merge_dictionaries(row.run_data, self.run_data)
                self.run_data = row.run_data
            # If run_data is not set in the assistant, use the database run_data
            if self.run_data is None and row.run_data is not None:
                self.run_data = row.run_data

        # Update user_data from the database
        if row.user_data is not None:
            # If user_data is set in the assistant, merge it with the database user_data.
            # The assistant user_data takes precedence
            if self.user_data is not None and row.user_data is not None:
                # Updates db_row.user_data with self.user_data
                merge_dictionaries(row.user_data, self.user_data)
                self.user_data = row.user_data
            # If user_data is not set in the assistant, use the database user_data
            if self.user_data is None and row.user_data is not None:
                self.user_data = row.user_data

        # Update task_data from the database
        if row.task_data is not None:
            # If task_data is set in the assistant, merge it with the database task_data.
            # The assistant task_data takes precedence
            if self.task_data is not None and row.task_data is not None:
                # Updates db_row.task_data with self.task_data
                merge_dictionaries(row.task_data, self.task_data)
                self.task_data = row.task_data
            # If task_data is not set in the assistant, use the database task_data
            if self.task_data is None and row.task_data is not None:
                self.task_data = row.task_data

    def read_from_storage(self) -> Optional[AssistantRun]:
        """Load the AssistantRun from storage"""

        if self.storage is not None and self.run_id is not None:
            self.db_row = self.storage.read(run_id=self.run_id)
            if self.user_id is not None and self.db_row is not None and self.db_row.user_id != self.user_id:
                logger.error(f"SECURITY ERROR: User id mismatch: {self.user_id} != {self.db_row.user_id}")
                return None
            if self.db_row is not None:
                logger.debug(f"-*- Loading run: {self.db_row.run_id}")
                self.from_database_row(row=self.db_row)
                logger.debug(f"-*- Loaded run: {self.run_id}")
        return self.db_row

    def write_to_storage(self) -> Optional[AssistantRun]:
        """Save the AssistantRun to the storage"""

        if self.storage is not None:
            self.db_row = self.storage.upsert(row=self.to_database_row())
        return self.db_row

    def add_introduction(self, introduction: str) -> None:
        """Add assistant introduction to the chat history"""

        if introduction is not None:
            if len(self.memory.chat_history) == 0:
                self.memory.add_chat_message(Message(role="assistant", content=introduction))

    def create_run(self) -> Optional[str]:
        """Create a run in the database and return the run_id.
        This function:
            - Creates a new run in the storage if it does not exist
            - Load the assistant from the storage if it exists
        """

        # If a database_row exists, return the id from the database_row
        if self.db_row is not None:
            return self.db_row.run_id

        # Create a new run or load an existing run
        if self.storage is not None:
            # Load existing run if it exists
            logger.debug(f"Reading run: {self.run_id}")
            self.read_from_storage()

            # Create a new run
            if self.db_row is None:
                logger.debug("-*- Creating new assistant run")
                if self.introduction:
                    self.add_introduction(self.introduction)
                self.db_row = self.write_to_storage()
                if self.db_row is None:
                    raise Exception("Failed to create new assistant run in storage")
                logger.debug(f"-*- Created assistant run: {self.db_row.run_id}")
                self.from_database_row(row=self.db_row)
                self._api_log_assistant_run()
        return self.run_id

    def _run(
        self, message: Optional[Union[List, Dict, str]] = None, stream: bool = True, **kwargs: Any
    ) -> Iterator[str]:
        logger.debug(f"*********** Run Start: {self.run_id} ***********")
        # Load run from storage
        self.read_from_storage()

        # Add a default LLMTask if tasks are empty
        _tasks = self.tasks
        if _tasks is None or len(_tasks) == 0:
            _tasks = [self.llm_task]

        # Metadata for all tasks in this run
        task_data: List[Dict[str, Any]] = []
        # Final LLM response after running all tasks
        run_output = ""

        # -*- Generate response by running tasks
        current_task: Optional[Task] = None
        for idx, task in enumerate(_tasks, start=1):
            logger.debug(f"*********** Task {idx} Start ***********")

            # Set previous_task and current_task
            previous_task = current_task
            if previous_task is not None and previous_task.show_output:
                if stream:
                    yield "\n\n"
                run_output += "\n\n"

            current_task = task

            # -*- Prepare input message for the current_task
            current_task_message: Optional[Union[List, Dict, str]] = None
            if previous_task and previous_task.output is not None:
                # Convert current_task_message to json if it is a BaseModel
                if issubclass(previous_task.output.__class__, BaseModel):
                    current_task_message = previous_task.output.model_dump_json(exclude_none=True, indent=2)
                else:
                    current_task_message = previous_task.output
            else:
                current_task_message = message

            # -*- Update Task
            # Add run state to the task
            current_task.run_id = self.run_id
            current_task.assistant_name = self.name
            current_task.assistant_memory = self.memory
            current_task.run_message = message
            current_task.run_task_data = task_data

            # Set output parsing off. This is handled by the run() function
            current_task.parse_output = False

            # -*- Update LLMTask
            if isinstance(current_task, LLMTask):
                # Update LLM
                if current_task.llm is None and self.llm is not None:
                    current_task.llm = self.llm.model_copy()

            # -*- Run Task
            if stream and current_task.streamable:
                for chunk in current_task.run(message=current_task_message, stream=True, **kwargs):
                    if current_task.show_output:
                        run_output += chunk if isinstance(chunk, str) else ""
                        yield chunk if isinstance(chunk, str) else ""
            else:
                current_task_response = current_task.run(message=current_task_message, stream=False, **kwargs)  # type: ignore
                current_task_response_str = ""
                try:
                    if current_task_response:
                        if isinstance(current_task_response, str):
                            current_task_response_str = current_task_response
                        elif issubclass(current_task_response.__class__, BaseModel):
                            current_task_response_str = current_task_response.model_dump_json(
                                exclude_none=True, indent=2
                            )
                        else:
                            current_task_response_str = json.dumps(current_task_response)

                        if current_task.show_output:
                            if stream:
                                yield current_task_response_str
                            else:
                                run_output += current_task_response_str
                except Exception as e:
                    logger.debug(f"Failed to convert task response to json: {e}")

            logger.debug(f"*********** Task {idx} End ***********")

        # -*- Save run to storage
        self.write_to_storage()

        # -*- Send run event for monitoring
        llm_response_type = "text"
        if self.output_model is not None:
            llm_response_type = "json"
        elif self.markdown:
            llm_response_type = "markdown"
        event_info = {
            "tasks": task_data,
        }
        event_data = {
            "user_message": message,
            "llm_response": run_output,
            "llm_response_type": llm_response_type,
            "info": event_info,
            "metrics": self.llm.metrics if self.llm else None,
        }
        self._api_log_assistant_event(event_type="run", event_data=event_data)

        # -*- Update run output
        self.output = run_output

        # -*- Yield final response if not streaming
        if not stream:
            yield run_output
        logger.debug(f"*********** Run End: {self.run_id} ***********")

    def run(
        self, message: Optional[Union[List, Dict, str]] = None, stream: bool = True, **kwargs: Any
    ) -> Union[Iterator[str], str, BaseModel]:
        # Convert response to structured output if output_model is set
        if self.output_model is not None and self.parse_output:
            logger.debug("Setting stream=False as output_model is set")
            json_resp = next(self._run(message=message, stream=False))
            try:
                structured_output = None
                if (
                    isinstance(self.output_model, str)
                    or isinstance(self.output_model, dict)
                    or isinstance(self.output_model, list)
                ):
                    structured_output = json.loads(json_resp)
                elif issubclass(self.output_model, BaseModel):
                    try:
                        structured_output = self.output_model.model_validate_json(json_resp)
                    except ValidationError:
                        # Check if response starts with ```json
                        if json_resp.startswith("```json"):
                            json_resp = json_resp.replace("```json\n", "").replace("\n```", "")
                            try:
                                structured_output = self.output_model.model_validate_json(json_resp)
                            except ValidationError as exc:
                                logger.warning(f"Failed to validate response: {exc}")

                # -*- Update assistant output to the structured output
                if structured_output is not None:
                    self.output = structured_output
            except Exception as e:
                logger.warning(f"Failed to convert response to output model: {e}")

            return self.output or json_resp
        else:
            if stream and self.streamable:
                resp = self._run(message=message, stream=True, **kwargs)
                return resp
            else:
                resp = self._run(message=message, stream=False, **kwargs)
                return next(resp)

    def chat(
        self, message: Union[List, Dict, str], stream: bool = True, **kwargs: Any
    ) -> Union[Iterator[str], str, BaseModel]:
        return self.run(message=message, stream=stream, **kwargs)

    def _chat_raw(
        self, messages: List[Message], user_message: Optional[str] = None, stream: bool = True
    ) -> Iterator[Dict]:
        logger.debug("*********** Assistant Chat Raw Start ***********")
        if self.llm is None:
            raise Exception("LLM not set")

        # Load run from storage
        self.read_from_storage()

        # -*- Add user message to the memory - this is added to the chat_history
        if user_message:
            self.memory.add_chat_message(Message(role="user", content=user_message))

        # -*- Generate response
        batch_llm_response_message = {}
        if stream:
            for response_delta in self.llm.generate_stream(messages=messages):
                yield response_delta
        else:
            batch_llm_response_message = self.llm.generate(messages=messages)

        # -*- Add prompts and response to the memory - these are added to the llm_messages
        self.memory.add_llm_messages(messages=messages)

        # Add llm response to the chat history
        # LLM Response is the last message in the messages list
        llm_response_message = messages[-1]
        try:
            self.memory.add_chat_message(llm_response_message)
        except Exception as e:
            logger.warning(f"Failed to add llm response to memory: {e}")

        # -*- Save run to storage
        self.write_to_storage()

        # -*- Send assistant event for monitoring
        event_data = {
            "user_message": user_message,
            "llm_response": llm_response_message,
            "messages": [m.model_dump(exclude_none=True) for m in messages],
            "metrics": self.llm.metrics,
        }
        self._api_log_assistant_event(event_type="chat_raw", event_data=event_data)

        # -*- Yield final response if not streaming
        if not stream:
            yield batch_llm_response_message
        logger.debug("*********** Assistant Chat Raw End ***********")

    def chat_raw(
        self, messages: List[Message], user_message: Optional[str] = None, stream: bool = True
    ) -> Union[Iterator[Dict], Dict]:
        if self.tasks and len(self.tasks) > 0:
            raise Exception("chat_raw does not support tasks")
        resp = self._chat_raw(messages=messages, user_message=user_message, stream=stream)
        if stream:
            return resp
        else:
            return next(resp)

    def rename(self, name: str) -> None:
        """Rename the assistant for the current run"""
        # -*- Read run to storage
        self.read_from_storage()
        # -*- Rename assistant
        self.name = name
        # -*- Save run to storage
        self.write_to_storage()
        # -*- Log assistant run
        self._api_log_assistant_run()

    def rename_run(self, name: str) -> None:
        """Rename the current run"""
        # -*- Read run to storage
        self.read_from_storage()
        # -*- Rename run
        self.run_name = name
        # -*- Save run to storage
        self.write_to_storage()
        # -*- Log assistant run
        self._api_log_assistant_run()

    def generate_name(self) -> str:
        """Generate a name for the run using the first 6 messages of the chat history"""
        if self.llm is None:
            raise Exception("LLM not set")

        _conv = "Conversation\n"
        _messages_for_generating_name = []
        try:
            if self.memory.chat_history[0].role == "assistant":
                _messages_for_generating_name = self.memory.chat_history[1:6]
            else:
                _messages_for_generating_name = self.memory.chat_history[:6]
        except Exception as e:
            logger.warning(f"Failed to generate name: {e}")
        finally:
            if len(_messages_for_generating_name) == 0:
                _messages_for_generating_name = self.memory.llm_messages[-4:]

        for message in _messages_for_generating_name:
            _conv += f"{message.role.upper()}: {message.content}\n"

        _conv += "\n\nConversation Name: "

        system_message = Message(
            role="system",
            content="Please provide a suitable name for this conversation in maximum 5 words. "
            "Remember, do not exceed 5 words.",
        )
        user_message = Message(role="user", content=_conv)
        generate_name_messages = [system_message, user_message]
        generated_name = self.llm.response(messages=generate_name_messages)
        if len(generated_name.split()) > 15:
            logger.error("Generated name is too long. Trying again.")
            return self.generate_name()
        return generated_name.replace('"', "").strip()

    def auto_rename_run(self) -> None:
        """Automatically rename the run"""
        # -*- Read run to storage
        self.read_from_storage()
        # -*- Generate name for run
        generated_name = self.generate_name()
        logger.debug(f"Generated name: {generated_name}")
        self.run_name = generated_name
        # -*- Save run to storage
        self.write_to_storage()
        # -*- Log assistant run
        self._api_log_assistant_run()

    ###########################################################################
    # Api functions
    ###########################################################################

    def _api_log_assistant_run(self):
        if not self.monitoring:
            return

        from phi.api.assistant import create_assistant_run, AssistantRunCreate

        try:
            database_row: AssistantRun = self.db_row or self.to_database_row()
            create_assistant_run(
                run=AssistantRunCreate(
                    run_id=database_row.run_id,
                    assistant_data=database_row.assistant_dict(),
                ),
            )
        except Exception as e:
            logger.debug(f"Could not create assistant monitor: {e}")

    def _api_log_assistant_event(self, event_type: str = "run", event_data: Optional[Dict[str, Any]] = None) -> None:
        if not self.monitoring:
            return

        from phi.api.assistant import create_assistant_event, AssistantEventCreate

        try:
            database_row: AssistantRun = self.db_row or self.to_database_row()
            create_assistant_event(
                event=AssistantEventCreate(
                    run_id=database_row.run_id,
                    assistant_data=database_row.assistant_dict(),
                    event_type=event_type,
                    event_data=event_data,
                ),
            )
        except Exception as e:
            logger.debug(f"Could not create assistant event: {e}")

    ###########################################################################
    # Print Response
    ###########################################################################

    def print_response(
        self,
        message: Optional[Union[List, Dict, str]] = None,
        stream: bool = True,
        markdown: bool = False,
        **kwargs: Any,
    ) -> None:
        from phi.cli.console import console
        from rich.live import Live
        from rich.table import Table
        from rich.status import Status
        from rich.progress import Progress, SpinnerColumn, TextColumn
        from rich.box import ROUNDED
        from rich.markdown import Markdown

        if markdown:
            self.markdown = True

        if self.output_model is not None:
            markdown = False
            self.markdown = False
            stream = False

        if stream:
            response = ""
            with Live() as live_log:
                status = Status("Working...", spinner="dots")
                live_log.update(status)
                response_timer = Timer()
                response_timer.start()
                for resp in self.run(message, stream=True, **kwargs):
                    if isinstance(resp, str):
                        response += resp
                    _response = Markdown(response) if self.markdown else response

                    table = Table(box=ROUNDED, border_style="blue", show_header=False)
                    if message:
                        table.show_header = True
                        table.add_column("Message")
                        table.add_column(get_text_from_message(message))
                    table.add_row(f"Response\n({response_timer.elapsed:.1f}s)", _response)  # type: ignore
                    live_log.update(table)
                response_timer.stop()
        else:
            response_timer = Timer()
            response_timer.start()
            with Progress(
                SpinnerColumn(spinner_name="dots"), TextColumn("{task.description}"), transient=True
            ) as progress:
                progress.add_task("Working...")
                response = self.run(message, stream=False, **kwargs)  # type: ignore

            response_timer.stop()
            _response = Markdown(response) if self.markdown else response

            table = Table(box=ROUNDED, border_style="blue", show_header=False)
            if message:
                table.show_header = True
                table.add_column("Message")
                table.add_column(get_text_from_message(message))
            table.add_row(f"Response\n({response_timer.elapsed:.1f}s)", _response)  # type: ignore
            console.print(table)

    def cli_app(
        self,
        user: str = "User",
        emoji: str = ":sunglasses:",
        stream: bool = True,
        markdown: bool = True,
        exit_on: Tuple[str, ...] = ("exit", "bye"),
    ) -> None:
        from rich.prompt import Prompt

        while True:
            message = Prompt.ask(f"[bold] {emoji} {user} [/bold]")
            if message in exit_on:
                break

            self.print_response(message=message, stream=stream, markdown=markdown)



from uuid import uuid4
from typing import List, Any, Optional, Dict, Union, Type, Iterator

from pydantic import BaseModel, ConfigDict, field_validator

from phi.memory.assistant import AssistantMemory
from phi.utils.log import logger, set_log_level_to_debug


class Task(BaseModel):
    # -*- Task settings
    # Task UUID (autogenerated if not set)
    task_id: Optional[str] = None
    # Task name
    task_name: Optional[str] = None

    # -*- Assistant state
    assistant_name: Optional[str] = None
    assistant_memory: Optional[AssistantMemory] = None

    # -*- Run state
    run_id: Optional[str] = None
    run_message: Optional[Union[List, Dict, str]] = None
    run_task_data: Optional[List[Dict[str, Any]]] = None

    # -*- Task Output Settings
    # Provide an output model for the responses
    output_model: Optional[Union[str, List, Type[BaseModel]]] = None
    # If True, the output is converted into the output_model (pydantic model or json dict)
    parse_output: bool = True
    # -*- Final output of this Task
    output: Optional[Any] = None
    # If True, shows the output of the task in the assistant.run()
    show_output: bool = True

    # If True, enable debug logs
    debug_mode: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("debug_mode", mode="before")
    def set_log_level(cls, v: bool) -> bool:
        if v:
            set_log_level_to_debug()
            logger.debug("Debug logs enabled")
        return v

    @property
    def streamable(self) -> bool:
        return False

    def set_task_id(self) -> None:
        if self.task_id is None:
            self.task_id = str(uuid4())

    def prepare_task(self) -> None:
        self.set_task_id()

    def run(
        self,
        message: Optional[Union[List, Dict, str]] = None,
        stream: bool = True,
        **kwargs: Dict[str, Any],
    ) -> Union[Iterator[str], str, BaseModel]:
        raise NotImplementedError

    def to_dict(self) -> Dict[str, Any]:
        _dict = {
            "task_id": self.task_id,
            "task_name": self.task_name,
            "output": self.output,
        }
        return _dict



from uuid import uuid4
from textwrap import dedent
from typing import List, Any, Optional, Dict, Iterator, Union

from pydantic import BaseModel, ConfigDict, field_validator, Field

from phi.assistant.assistant import Assistant
from phi.tools.function import Function
from phi.utils.log import logger, set_log_level_to_debug
from phi.utils.message import get_text_from_message
from phi.utils.timer import Timer


class Team(BaseModel):
    name: Optional[str] = None

    lead: Optional[Assistant] = None
    assistants: Optional[List[Assistant]] = None
    reviewer: Optional[Assistant] = None

    # -*- Run settings
    # Run UUID (autogenerated if not set)
    run_id: Optional[str] = Field(None, validate_default=True)
    # Run name
    run_name: Optional[str] = None
    # Metadata associated with this run
    run_data: Optional[Dict[str, Any]] = None

    # If markdown=true, add instructions to format the output using markdown
    markdown: bool = False

    # debug_mode=True enables debug logs
    debug_mode: bool = False
    # monitoring=True logs Assistant runs on phidata.com
    monitoring: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("debug_mode", mode="before")
    def set_log_level(cls, v: bool) -> bool:
        if v:
            set_log_level_to_debug()
            logger.debug("Debug logs enabled")
        return v

    @field_validator("run_id", mode="before")
    def set_run_id(cls, v: Optional[str]) -> str:
        return v if v is not None else str(uuid4())

    @property
    def streamable(self) -> bool:
        return True

    def assistant_delegation_function(self, assistant: Assistant, index: int) -> Function:
        def _delegate_task_to_assistant(task_description: str) -> str:
            return assistant.run(task_description, stream=False)  # type: ignore

        assistant_name = assistant.name.replace(" ", "_").lower() if assistant.name else f"assistant_{index}"
        delegation_function = Function.from_callable(_delegate_task_to_assistant)
        delegation_function.name = f"delegate_task_to_{assistant_name}"
        delegation_function.description = dedent(
            f"""Use this function to delegate a task to {assistant_name}
        Args:
            task_description (str): A clear and concise description of the task the assistant should achieve.
        Returns:
            str: The result of the delegated task.
        """
        )
        return delegation_function

    @property
    def leader(self) -> Assistant:
        """Returns the team leader"""

        if self.lead:
            return self.lead

        _delegation_functions = []

        _system_prompt = ""
        if self.assistants and len(self.assistants) > 0:
            _system_prompt += "You are the leader of a team of AI Assistants "
            if self.name:
                _system_prompt += f"called '{self.name}'"
        else:
            _system_prompt += "You are an AI Assistant"

        _system_prompt += "and your goal is to respond to the users message in the best way possible. "
        _system_prompt += "This is an important task and must be done with correctly.\n\n"

        if self.assistants and len(self.assistants) > 0:
            _system_prompt += (
                "Given a user message you can respond directly or delegate tasks to the following assistants depending on their role "
                "and the tools available to them. "
            )
            _system_prompt += "\n\n<assistants>"
            for assistant_index, assistant in enumerate(self.assistants):
                _system_prompt += f"\nAssistant {assistant_index+1}:\n"
                if assistant.name:
                    _system_prompt += f"Name: {assistant.name}\n"
                if assistant.role:
                    _system_prompt += f"Role: {assistant.role}\n"
                if assistant.tools is not None:
                    _tools = []
                    for _tool in assistant.tools:
                        if callable(_tool):
                            _tools.append(_tool.__name__)
                    _system_prompt += f"Available tools: {', '.join(_tools)}\n"
                _delegation_functions.append(self.assistant_delegation_function(assistant, assistant_index))
            _system_prompt += "</assistants>\n"

        if self.reviewer is None:
            _system_prompt += "You must always review the responses from the assistants and re-run tasks if the result is not satisfactory."

        return Assistant(
            system_prompt=_system_prompt,
            tools=_delegation_functions,
        )

    def _run(
        self, message: Optional[Union[List, Dict, str]] = None, stream: bool = True, **kwargs: Any
    ) -> Iterator[str]:
        logger.debug(f"*********** Team Run Start: {self.run_id} ***********")

        # Get the team leader
        leader = self.leader

        # Final LLM response after running all tasks
        run_output = ""

        if stream and leader.streamable:
            for chunk in leader.run(message=message, stream=True, **kwargs):
                run_output += chunk if isinstance(chunk, str) else ""
                yield chunk if isinstance(chunk, str) else ""
            yield "\n\n"
            run_output += "\n\n"
        else:
            try:
                leader_response = leader.run(message=message, stream=False, **kwargs)
                if stream:
                    yield leader_response  # type: ignore
                    yield "\n\n"
                else:
                    run_output += leader_response  # type: ignore
                    run_output += "\n\n"
            except Exception as e:
                logger.debug(f"Failed to convert task response to json: {e}")

        if not stream:
            yield run_output
        logger.debug(f"*********** Team Run End: {self.run_id} ***********")

    def run(
        self, message: Optional[Union[List, Dict, str]] = None, stream: bool = True, **kwargs: Any
    ) -> Union[Iterator[str], str, BaseModel]:
        if stream and self.streamable:
            resp = self._run(message=message, stream=True, **kwargs)
            return resp
        else:
            resp = self._run(message=message, stream=False, **kwargs)
            return next(resp)

    def print_response(
        self,
        message: Optional[Union[List, Dict, str]] = None,
        stream: bool = True,
        markdown: bool = False,
        **kwargs: Any,
    ) -> None:
        from phi.cli.console import console
        from rich.live import Live
        from rich.table import Table
        from rich.status import Status
        from rich.progress import Progress, SpinnerColumn, TextColumn
        from rich.box import ROUNDED
        from rich.markdown import Markdown

        if markdown:
            self.markdown = True

        if stream:
            response = ""
            with Live() as live_log:
                status = Status("Working...", spinner="dots")
                live_log.update(status)
                response_timer = Timer()
                response_timer.start()
                for resp in self.run(message, stream=True, **kwargs):
                    response += resp if isinstance(resp, str) else ""
                    _response = response if not markdown else Markdown(response)

                    table = Table(box=ROUNDED, border_style="blue", show_header=False)
                    if message:
                        table.show_header = True
                        table.add_column("Message")
                        table.add_column(get_text_from_message(message))
                    table.add_row(f"Response\n({response_timer.elapsed:.1f}s)", _response)  # type: ignore
                    live_log.update(table)
                response_timer.stop()
        else:
            response_timer = Timer()
            response_timer.start()
            with Progress(
                SpinnerColumn(spinner_name="dots"), TextColumn("{task.description}"), transient=True
            ) as progress:
                progress.add_task("Working...")
                response = self.run(message, stream=False, **kwargs)  # type: ignore

            response_timer.stop()
            _response = response if not markdown else Markdown(response)

            table = Table(box=ROUNDED, border_style="blue", show_header=False)
            if message:
                table.show_header = True
                table.add_column("Message")
                table.add_column(get_text_from_message(message))
            table.add_row(f"Response\n({response_timer.elapsed:.1f}s)", _response)  # type: ignore
            console.print(table)
